# Function to build a function to put inside compute
build_compute_workflow <- function(recipe) {
  ## TODO: Sort out system and order
  ## Alternatively, this function could just be defined by the user
  # Define function
  ## TODO: Define returns
  my_compute_function = function(recipe, hcst, obs, fcst = NULL) {
    modules <- tolower(strsplit(recipe$Run$startR_workflow$modules,
                                ", | |,")[[1]])
    data <- list(hcst = hcst, obs = obs, fcst = fcst)
    data <- preprocess_datasets(recipe, data)
    skill_metrics <- NULL
    for (module in modules) {
      if (module == "calibration") {
        data <- calibrate_datasets(recipe, data)
      } else if (module == "anomalies") {
        data <- compute_anomalies(recipe, data)
      } else if (module == "skill") {
        skill_metrics <- compute_skill_metrics(recipe, data)
      }
    }
    ## TODO: Define what to return depending the modules called + the recipe
    # Eliminate NULL elements from the return list
    return_list <- list(hcst = data$hcst,
                        obs = data$obs,
                        fcst = data$fcst,
                        skill = skill_metrics)
    return_list <- return_list[!sapply(return_list, is.null)]
    return(return_list)
  }
  info(recipe$Run$logger, "##### COMPUTE FUNCTION DEFINED #####")
  return(my_compute_function)
}

convert_to_s2dv_cube <- function(new_cube, original_cube) {
  ## TODO: Make sure all the metadata is created correctly
  attr(new_cube, "Variables") <- attr(original_cube, "Variables")
  attr(new_cube, "FileSelectors") <- attr(original_cube, "FileSelectors")

  class(new_cube) <- "startR_array"
  new_cube <- as.s2dv_cube(new_cube)
  return(new_cube)
}

run_compute_workflow <- function(recipe, data) {

  modules <- tolower(strsplit(recipe$Run$startR_workflow$modules,
                              ", | |,")[[1]])
  # Step 1: Define the function
  my_compute_function <- build_compute_workflow(recipe)
  
  # ---------------------------------------------------------------------------
  # Step 2: Define the inputs and outputs for Compute()
  # Create target dimensions
  target_dims <- list(hcst = c('sday', 'sweek', 'syear', 'ensemble'),
                      obs = c('sday', 'sweek', 'syear'))
  inputs <- list(hcst = data$hcst, obs = data$obs)
  if (!is.null(data$fcst)) {
    target_dims <- c(target_dims,
                     list(fcst = c('sday', 'sweek', 'syear', 'ensemble')))
    inputs <- c(inputs, list(fcst = data$fcst))
  }
  # Create output dimensions
  ## TODO: Add several conditions depending on user requests
  output_dims <- list(hcst = c('sday', 'sweek', 'syear', 'ensemble'),
                      obs = c('sday', 'sweek', 'syear', 'ensemble'))
  if (!is.null(data$fcst)) {
    output_dims <- c(output_dims,
                     list(fcst = c('sday', 'sweek', 'syear', 'ensemble')))
  }
  if ("skill" %in% modules) {
    output_dims <- c(output_dims,
                     list(skill = c('metric', 'syear')))
  }

  # ---------------------------------------------------------------------------
  # Step 3: Generate the Step and call Compute()
  ## TODO: Add use attributes
  step <- Step(fun = my_compute_function,
               target_dims = target_dims,
               output_dims = output_dims)
  wf <- AddStep(inputs = inputs,
                step = step,
                recipe = recipe)
  res <- Compute(wf$hcst,
                 chunks = recipe$Run$startR_workflow$chunk_along)

  info(recipe$Run$logger,
       "##### COMPUTE SECTION ENDED, REFORMATTING OUTPUT #####")

  # ---------------------------------------------------------------------------
  # Step 4: Convert results to s2dv_cube/list format
  result <- list()
  for (cube in c("hcst", "obs", "fcst")) {
    if (!is.null(res[[cube]])) {
      result[[cube]] <- convert_to_s2dv_cube(res[[cube]], data[[cube]])
    } else {
      result[[cube]] <- NULL
    }
  }
  ## TODO: Transform skill metrics back into a NAMED list. To do this, Apply()
  ## should return the attributes of the skill array so that we have the name
  ## of each metric, or we should find an alternative way to retrieve them.
  if (!is.null(res$skill)) {
    metric_list <- c("rpss", "rpss_significance") ## to be removed
    result$skill <- .drop_dims(res$skill)
    # result$skill_metrics <- apply(seq(dim(skill)[1]), function(x) {res$skill[x, , , ]})
  }
  info(recipe$Run$logger,
       "##### DATA RETURNED AS A NAMED LIST #####")
  return(result)
}

